'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var backendCommon = require('@backstage/backend-common');
var Router = require('express-promise-router');
var DataLoader = require('dataloader');
var graphql = require('graphql');
var helmet = require('helmet');
var catalogClient = require('@backstage/catalog-client');
var graphqlYoga = require('graphql-yoga');
var graphqlModules$1 = require('@envelop/graphql-modules');
var dataloader = require('@envelop/dataloader');
var graphqlModules = require('graphql-modules');
var schema = require('@graphql-tools/schema');
var loadFiles = require('@graphql-tools/load-files');
var GraphQLJSON = require('graphql-type-json');
var catalogModel = require('@backstage/catalog-model');
var lodash = require('lodash');
var graphqlRelay = require('graphql-relay');
var utils = require('@graphql-tools/utils');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var Router__default = /*#__PURE__*/_interopDefaultLegacy(Router);
var DataLoader__default = /*#__PURE__*/_interopDefaultLegacy(DataLoader);
var helmet__default = /*#__PURE__*/_interopDefaultLegacy(helmet);
var GraphQLJSON__default = /*#__PURE__*/_interopDefaultLegacy(GraphQLJSON);

function createLoader({ catalog }, options) {
  return new DataLoader__default["default"](
    async (entityRefs) => {
      const result = await catalog.getEntitiesByRefs({
        entityRefs
      });
      return result.items.map(
        (entity, index) => entity != null ? entity : new graphql.GraphQLError(`no such node with ref: '${entityRefs[index]}'`)
      );
    },
    options
  );
}

const refToId = (ref) => {
  return typeof ref === "string" ? ref : catalogModel.stringifyEntityRef(ref);
};

const Catalog = graphqlModules.createModule({
  id: "catalog",
  typeDefs: loadFiles.loadFilesSync(backendCommon.resolvePackagePath("@frontside/backstage-plugin-graphql", "src/app/modules/catalog/catalog.graphql")),
  resolvers: {
    Lifecycle: {
      EXPERIMENTAL: "experimental",
      PRODUCTION: "production",
      DEPRECATED: "deprecated"
    },
    JSON: GraphQLJSON__default["default"],
    JSONObject: GraphQLJSON.GraphQLJSONObject,
    Query: {
      entity: (_, { name, kind, namespace = "default" }, { refToId: refToId$1 = refToId }) => ({ id: refToId$1({ name, kind, namespace }) })
    }
  }
});

const Core = graphqlModules.createModule({
  id: "core",
  typeDefs: loadFiles.loadFilesSync(backendCommon.resolvePackagePath("@frontside/backstage-plugin-graphql", "src/app/modules/core/core.graphql")),
  resolvers: {
    Node: {
      id: async ({ id }, _, { loader }) => {
        const entity = await loader.load(id);
        if (!entity)
          return null;
        return id;
      }
    },
    Query: { node: (_, { id }) => ({ id }) }
  }
});

function getObjectTypeName(iface, directive) {
  if (directive && "generatedTypeName" in directive)
    return directive.generatedTypeName;
  return iface.name.slice(1);
}
function filterEntityRefs(entity, relationType, targetKind) {
  var _a, _b;
  return (_b = (_a = entity == null ? void 0 : entity.relations) == null ? void 0 : _a.filter(({ type }) => !relationType || type === relationType).flatMap(({ targetRef }) => {
    const ref = catalogModel.parseEntityRef(targetRef);
    return !targetKind || ref.kind.toLowerCase() === targetKind.toLowerCase() ? [ref] : [];
  })) != null ? _b : [];
}
function isConnectionType(type) {
  return graphql.isInterfaceType(type) && type.name === "Connection" || graphql.isNonNullType(type) && isConnectionType(type.ofType);
}
function createConnectionType(typeName, fieldType, nodeType) {
  const wrappedEdgeType = fieldType.getFields().edges.type;
  const edgeType = wrappedEdgeType.ofType.ofType.ofType;
  return new graphql.GraphQLObjectType({
    name: `${typeName}Connection`,
    fields: {
      ...fieldType.toConfig().fields,
      edges: {
        type: new graphql.GraphQLNonNull(new graphql.GraphQLList(new graphql.GraphQLNonNull(new graphql.GraphQLObjectType({
          name: `${typeName}Edge`,
          fields: {
            ...edgeType.toConfig().fields,
            node: {
              type: new graphql.GraphQLNonNull(nodeType)
            }
          },
          interfaces: [edgeType]
        }))))
      }
    },
    interfaces: [fieldType]
  });
}
function transformDirectives(sourceSchema, { logger = console } = {}) {
  const inheritsWithoutArgs = /* @__PURE__ */ new Set();
  const resolversMap = {};
  const typesToAdd = /* @__PURE__ */ new Map();
  const additionalInterfaces = {};
  function handleFieldDirective(field, directive) {
    if ("at" in directive && typeof directive.at !== "string" && (!Array.isArray(directive.at) || directive.at.some((a) => typeof a !== "string"))) {
      throw new Error(`The "at" argument of @field directive must be a string or an array of strings`);
    }
    field.resolve = async ({ id }, _, { loader }) => {
      var _a;
      const entity = await loader.load(id);
      if (!entity)
        return null;
      return (_a = lodash.get(entity, directive.at)) != null ? _a : directive.default;
    };
  }
  function handleRelationDirective(field, directive, schema) {
    var _a;
    const fieldType = field.type;
    if (graphql.isListType(fieldType) && isConnectionType(fieldType.ofType) || graphql.isNonNullType(fieldType) && graphql.isListType(fieldType.ofType) && isConnectionType(fieldType.ofType.ofType)) {
      throw new Error(`It's not possible to use a list of Connection type. Use either Connection type or list of specific type`);
    }
    const isList = graphql.isListType(fieldType) || graphql.isNonNullType(fieldType) && graphql.isListType(fieldType.ofType);
    if (isConnectionType(fieldType)) {
      if (directive.nodeType) {
        const nodeType = schema.getType(directive.nodeType);
        if (!nodeType) {
          throw new Error(`The interface "${directive.nodeType}" is not defined in the schema.`);
        }
        if (graphql.isInputType(nodeType)) {
          throw new Error(`The interface "${directive.nodeType}" is an input type and can't be used in a Connection.`);
        }
        if (graphql.isUnionType(nodeType)) {
          const resolveType = nodeType.resolveType;
          if (resolveType)
            logger.warn(
              `The "resolveType" function has already been implemented for "${nodeType.name}" union which may lead to undefined behavior`
            );
          const iface = (_a = typesToAdd.get(directive.nodeType)) != null ? _a : new graphql.GraphQLInterfaceType({
            name: directive.nodeType,
            interfaces: [schema.getType("Node")],
            fields: { id: { type: new graphql.GraphQLNonNull(graphql.GraphQLID) } },
            resolveType: (...args2) => {
              var _a2;
              return (_a2 = resolveType == null ? void 0 : resolveType(...args2)) != null ? _a2 : resolversMap.Node(...args2);
            }
          });
          typesToAdd.set(directive.nodeType, iface);
          nodeType.getTypes().forEach((type) => {
            var _a2;
            additionalInterfaces[type.name] = ((_a2 = additionalInterfaces[type.name]) != null ? _a2 : /* @__PURE__ */ new Set()).add(iface);
          });
          field.type = createConnectionType(nodeType.name, fieldType, iface);
        } else {
          const typeName = graphql.isInterfaceType(nodeType) ? getObjectTypeName(nodeType) : nodeType.name;
          field.type = createConnectionType(typeName, fieldType, nodeType);
        }
      }
      const mandatoryArgs = [
        ["first", "Int"],
        ["after", "String"],
        ["last", "Int"],
        ["before", "String"]
      ];
      const args = { ...field.args };
      mandatoryArgs.forEach(([name, type]) => {
        if (name in args) {
          const argType = args[name].type;
          if ((graphql.isNonNullType(argType) ? argType.ofType.toString() : argType.name) !== type) {
            throw new Error(`The field has mandatory argument "${name}" with different type than expected. Expected: ${type}`);
          }
        }
        args[name] = { type: type === "Int" ? graphql.GraphQLInt : graphql.GraphQLString };
      });
      field.args = args;
      field.resolve = async ({ id }, args2, { loader, refToId: refToId$1 = refToId }) => {
        const ids = filterEntityRefs(await loader.load(id), directive.name, directive.kind).map((ref) => ({ id: refToId$1(ref) }));
        return {
          ...graphqlRelay.connectionFromArray(ids, args2),
          count: ids.length
        };
      };
    } else {
      field.resolve = async ({ id }, _, { loader, refToId: refToId$1 = refToId }) => {
        var _a2;
        const ids = filterEntityRefs(await loader.load(id), directive.name, directive.kind).map((ref) => ({ id: refToId$1(ref) }));
        return isList ? ids : (_a2 = ids[0]) != null ? _a2 : null;
      };
    }
  }
  function validateExtendDirective(interfaceType, directive, schema) {
    var _a;
    if (!/^I[A-Z].*/.test(interfaceType.name)) {
      if (!("generatedTypeName" in directive)) {
        throw new Error(
          `The interface name "${interfaceType.name}" should started from capitalized 'I', like: "I${interfaceType.name}"`
        );
      }
      if (schema.getType(directive.generatedTypeName)) {
        throw new Error(
          `The type "${directive.generatedTypeName}" described in the @inherit directive is already declared in the schema`
        );
      }
    }
    if ("when" in directive !== "is" in directive) {
      throw new Error(
        `The @inherit directive for "${interfaceType.name}" should have both "when" and "is" arguments or none of them`
      );
    }
    if (!("when" in directive)) {
      if ("interface" in directive && inheritsWithoutArgs.has(directive.interface)) {
        throw new Error(
          `The @inherit directive of "${directive.interface}" without "when" and "is" arguments could be used only once`
        );
      } else {
        inheritsWithoutArgs.add(directive.interface);
      }
      const parentType = schema.getType(directive.interface);
      if (parentType) {
        const [inheritDirective] = (_a = utils.getDirective(schema, parentType, "inherit")) != null ? _a : [];
        if (inheritDirective && "interface" in inheritDirective && !("when" in inheritDirective) && Object.values(interfaceType.getFields()).some(
          (field) => graphql.isNonNullType(field.type)
        )) {
          throw new Error(
            `The interface "${interfaceType.name}" has required fields and can't be inherited from "${directive.interface}" without "when" and "is" arguments, because "${directive.interface}" has already been inherited without them`
          );
        }
      }
    }
    if ("when" in directive && (typeof directive.when !== "string" || Array.isArray(directive.when) && directive.when.some((a) => typeof a !== "string"))) {
      throw new Error(
        `The "when" argument of @inherit directive must be a string or an array of strings`
      );
    }
  }
  function defineResolver(iface, directive, schema) {
    var _a, _b;
    const objectType = getObjectTypeName(iface, directive);
    if (!resolversMap[iface.name])
      resolversMap[iface.name] = () => objectType;
    const inheritedInterface = schema.getType(directive.interface);
    if (!inheritedInterface)
      return;
    const [inheritDirective] = (_a = utils.getDirective(schema, inheritedInterface, "inherit")) != null ? _a : [];
    const inheritedObjectType = getObjectTypeName(
      inheritedInterface,
      inheritDirective
    );
    const resolveType = (_b = resolversMap[inheritedInterface.name]) != null ? _b : inheritedInterface.name === "Node" ? () => void 0 : () => inheritedObjectType;
    resolversMap[inheritedInterface.name] = async (source, context, info, abstractType) => {
      var _a2, _b2, _c, _d, _e;
      if ("when" in directive && "is" in directive) {
        const { id } = source;
        const { loader } = context;
        const entity = await loader.load(id);
        if (!entity)
          return void 0;
        if (lodash.isEqual(lodash.get(entity, directive.when), directive.is)) {
          return (_b2 = (_a2 = resolversMap[iface.name]) == null ? void 0 : _a2.call(resolversMap, source, context, info, abstractType)) != null ? _b2 : void 0;
        }
        return (_c = resolveType(source, context, info, abstractType)) != null ? _c : void 0;
      }
      return (_e = (_d = resolversMap[iface.name]) == null ? void 0 : _d.call(resolversMap, source, context, info, abstractType)) != null ? _e : void 0;
    };
  }
  function mapCompositeField(fieldConfig, fieldName, typeName, schema) {
    var _a, _b;
    const [fieldDirective] = (_a = utils.getDirective(schema, fieldConfig, "field")) != null ? _a : [];
    const [relationDirective] = (_b = utils.getDirective(schema, fieldConfig, "relation")) != null ? _b : [];
    if (fieldDirective && relationDirective) {
      throw new Error(
        `The field "${fieldName}" of "${typeName}" type has both @field and @relation directives at the same time`
      );
    }
    try {
      if (fieldDirective) {
        handleFieldDirective(fieldConfig, fieldDirective);
      } else if (relationDirective) {
        handleRelationDirective(fieldConfig, relationDirective, schema);
      }
    } catch (error) {
      const errorMessage = error instanceof Error ? error.message : error;
      throw new Error(
        `Error while processing directives on field "${fieldName}" of "${typeName}":
${errorMessage}`
      );
    }
    return fieldConfig;
  }
  function mapInterfaceType(interfaceType, schema) {
    var _a, _b, _c;
    const [inheritDirective] = (_a = utils.getDirective(schema, interfaceType, "inherit")) != null ? _a : [];
    if (!inheritDirective)
      return interfaceType;
    validateExtendDirective(interfaceType, inheritDirective, schema);
    defineResolver(interfaceType, inheritDirective, schema);
    const objectType = getObjectTypeName(interfaceType, inheritDirective);
    const inheritedInterfaces = traverseExtends(interfaceType, schema);
    const interfaces = [
      ...new Map(
        [
          ...(_c = (_b = additionalInterfaces[interfaceType.name]) == null ? void 0 : _b.values()) != null ? _c : [],
          ...inheritedInterfaces.flatMap((iface) => {
            var _a2, _b2;
            return [
              ...(_b2 = (_a2 = additionalInterfaces[iface.name]) == null ? void 0 : _a2.values()) != null ? _b2 : []
            ];
          }),
          ...inheritedInterfaces
        ].map((iface) => [iface.name, iface])
      ).values()
    ];
    const fields = [...interfaces].reverse().reduce(
      (acc, type) => ({ ...acc, ...type.toConfig().fields }),
      {}
    );
    const { astNode, extensionASTNodes, ...typeConfig } = interfaceType.toConfig();
    typeConfig.fields = fields;
    typesToAdd.set(
      objectType,
      new graphql.GraphQLObjectType({
        ...typeConfig,
        name: objectType,
        interfaces
      })
    );
    return new graphql.GraphQLInterfaceType({
      ...typeConfig,
      resolveType: (...args) => resolversMap[interfaceType.name](...args),
      interfaces: interfaces.filter((iface) => iface.name !== interfaceType.name)
    });
  }
  function mapUnionType(unionType, schema) {
    const typeConfig = unionType.toConfig();
    let hasInterfacesFromResolversMap = false;
    typeConfig.types = typeConfig.types.flatMap((type) => {
      if (graphql.isInterfaceType(type) && type.name in resolversMap) {
        hasInterfacesFromResolversMap = true;
        return utils.getImplementingTypes(
          type.name,
          schema
        ).map((name) => schema.getType(name));
      }
      return [type];
    });
    if (!hasInterfacesFromResolversMap)
      return unionType;
    const resolveType = typeConfig.resolveType;
    if (resolveType)
      logger.warn(
        `The "resolveType" function has already been implemented for "${unionType.name}" union which may lead to undefined behavior`
      );
    typeConfig.resolveType = (...args) => {
      var _a;
      return (_a = resolveType == null ? void 0 : resolveType(...args)) != null ? _a : resolversMap.Node(...args);
    };
    return new graphql.GraphQLUnionType(typeConfig);
  }
  const finalSchema = utils.mapSchema(utils.addTypes(utils.mapSchema(utils.mapSchema(sourceSchema, {
    [utils.MapperKind.COMPOSITE_FIELD]: mapCompositeField
  }), {
    [utils.MapperKind.INTERFACE_TYPE]: mapInterfaceType
  }), [...typesToAdd.values()]), {
    [utils.MapperKind.UNION_TYPE]: mapUnionType
  });
  if ("Node" in resolversMap) {
    const nodeType = finalSchema.getType("Node");
    const resolveType = nodeType.resolveType;
    if (resolveType)
      logger.warn(
        `The "resolveType" function has already been implemented for "Node" interface which may lead to undefined behavior`
      );
    nodeType.resolveType = (...args) => {
      var _a;
      return (_a = resolveType == null ? void 0 : resolveType(...args)) != null ? _a : resolversMap.Node(...args);
    };
  }
  return finalSchema;
}
function traverseExtends(type, schema) {
  var _a;
  const [inheritDirective] = (_a = utils.getDirective(schema, type, "inherit")) != null ? _a : [];
  const interfaces = [
    type,
    ...type.getInterfaces().flatMap((iface) => traverseExtends(iface, schema))
  ];
  if (inheritDirective && "interface" in inheritDirective) {
    const inheritedInterface = schema.getType(inheritDirective.interface);
    if (!graphql.isInterfaceType(inheritedInterface)) {
      throw new Error(
        `The interface "${inheritDirective.interface}" described in @inherit directive for "${type.name}" isn't abstract type or doesn't exist`
      );
    }
    if (interfaces.includes(inheritedInterface)) {
      throw new Error(
        `The interface "${inheritDirective.interface}" described in @inherit directive for "${type.name}" is already implemented by the type`
      );
    }
    interfaces.push(...traverseExtends(inheritedInterface, schema));
  }
  return interfaces;
}

function createGraphQLApp(options) {
  const { modules } = options;
  return graphqlModules.createApplication({
    schemaBuilder: (input) => transformDirectives(schema.makeExecutableSchema(input), options),
    modules: [Core, Catalog, ...modules != null ? modules : []]
  });
}

async function createRouter({
  config,
  logger,
  modules,
  plugins,
  createLoader: createLoader$1,
  refToId
}) {
  let yoga = null;
  const router = Router__default["default"]();
  const discovery = backendCommon.SingleHostDiscovery.fromConfig(config);
  const catalog = new catalogClient.CatalogClient({ discoveryApi: discovery });
  const application = createGraphQLApp({ modules, logger });
  router.get("/health", (_, response) => {
    response.json({ status: "ok" });
  });
  router.get("/schema", (_, response) => {
    response.send(graphql.printSchema(application.schema));
  });
  if (process.env.NODE_ENV === "development")
    router.use(
      helmet__default["default"].contentSecurityPolicy({
        directives: {
          defaultSrc: ["'self'", "'unsafe-inline'", "http://*"],
          scriptSrc: ["'self'", "'unsafe-inline'", "https://*"],
          imgSrc: ["'self'", "https: data:"]
        }
      })
    );
  router.use((req, res, next) => {
    if (!yoga) {
      yoga = graphqlYoga.createYoga({
        plugins: [
          graphqlModules$1.useGraphQLModules(application),
          dataloader.useDataLoader("loader", createLoader$1 != null ? createLoader$1 : createLoader),
          ...plugins != null ? plugins : []
        ],
        context: { application, catalog, refToId },
        logging: logger,
        graphqlEndpoint: req.baseUrl
      });
    }
    return yoga(req, res, next);
  });
  router.use(backendCommon.errorHandler());
  return router;
}

exports.createRouter = createRouter;
//# sourceMappingURL=index.cjs.js.map
